@@ -27,11 +27,17 @@ extern "C" {
 #include <stddef.h>
 #include <stdint.h>
 
+#ifdef JUICE_HAS_EXPORT_HEADER
+#include "juice_export.h"
+#endif
+
+#ifndef JUICE_EXPORT
 #ifdef _WIN32
 #define JUICE_EXPORT __declspec(dllexport)
 #else
 #define JUICE_EXPORT
 #endif
+#endif
 
 #define JUICE_ERR_SUCCESS 0
 #define JUICE_ERR_INVALID -1   // invalid argument
@@ -40,7 +46,7 @@ extern "C" {
 
 // ICE Agent
 
-#define JUICE_MAX_ADDRESS_STRING_LEN 56
+#define JUICE_MAX_ADDRESS_STRING_LEN 64
 #define JUICE_MAX_CANDIDATE_SDP_STRING_LEN 256
 #define JUICE_MAX_SDP_STRING_LEN 4096
 
